""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
from handlerworkers.apikeys import CSAPIKeys
from handlerworkers.cluster import add_cluster_command, get_app_keys
from handlerworkers.cluster import save_app_keys_for_new_nodes, save_encryption_key_for_new_nodes

from os import makedirs
import argparse
from os.path import join,isfile
from utilities.util import write_to_file
import sys
import random, string
from OpenSSL import crypto
from cshmac.requests import HmacAuth
from handlerworkers.userinfo import CSUserInfo
import urllib3
import requests
import json
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

PLAYBOOK_APPLIANCE_UUID = '6f3626c1-0da7-4b44-8d5e-43750f780a47'



APPLIANCE_PUBLIC_KEY_PATH='/opt/cyops/configs/keys/APPLIANCE_PUBLIC_KEY'
APPLIANCE_PRIVATE_KEY_PATH='/opt/cyops/configs/keys/APPLIANCE_PRIVATE_KEY'

SSO_DEFAULT_CERT = """-----BEGIN CERTIFICATE-----
MIIFtTCCA52gAwIBAgIJAOihHrl3izm8MA0GCSqGSIb3DQEBBQUAMEUxCzAJBgNV
BAYTAkFVMRMwEQYDVQQIEwpTb21lLVN0YXRlMSEwHwYDVQQKExhJbnRlcm5ldCBX
aWRnaXRzIFB0eSBMdGQwHhcNMTcwNDEwMTEyNTA4WhcNMjAwNDA5MTEyNTA4WjBF
MQswCQYDVQQGEwJBVTETMBEGA1UECBMKU29tZS1TdGF0ZTEhMB8GA1UEChMYSW50
ZXJuZXQgV2lkZ2l0cyBQdHkgTHRkMIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIIC
CgKCAgEAzB0AC+G/emNtH11J7Juo+3kVihpkfsMhxpyKB61n48n3FMeTkV9DESEJ
r4DBUpGidntGk4gyLKLqlApJiYs1lUps2vW211/aKNy5IUo+088bwO5NUcvv4Spi
5dbVW1cv+NFIJQh7VbgZqAKJvDmu3Rl24maNbTw0SfNLcVLKouckIUG2vT3Gh1Fd
dFjc+7vXCUQcFn5UzDUUGoV1yxLwD8vgpuWueHC6L1PNptGuXv6K49zpBANVRyM0
03QGaBXronzNyZ8BBCThv6zm4vLHhQ4cm3CEYWLKbj85DfLCDtegg/R/aEMq1A2k
MdXMnf9Pq0MdooOy92IcPvjkfL/lhnzgfTeE1Q5vhbGs9Bi/7yYF9pH/E/AT9FPS
dNifHpGkeN0E8J/WuK9vDcOEibfffkGlBivCHDIeIHScLjVm7wJrxsMEc/Cn9DoF
J7f+hYzvjtuuAWXcBRmT0EIB+gkSPqTwK6z1iH+JjuoUq5grdOaXwDqFQyLTYiP8
ZJ81Fn8G7X0RtEyOXKCF0yjGGKc+YenSuB2jyT9TswJRuoGjlNGpAaAaG6zaiP3g
YuQ5LDXyxpMFc7ZCQHChqTE1rrQ8d8uzeqYhu1VZbilgLlxnh9BqzahcZGmToMQy
fZiVuE95xGaZ4pcaCsGL9/TAWL8hejD+c1Hgpb6z2Lz4Y+knJmcCAwEAAaOBpzCB
pDAdBgNVHQ4EFgQUgOc6PAKBMuJiGP/P98onYOYjUbUwdQYDVR0jBG4wbIAUgOc6
PAKBMuJiGP/P98onYOYjUbWhSaRHMEUxCzAJBgNVBAYTAkFVMRMwEQYDVQQIEwpT
b21lLVN0YXRlMSEwHwYDVQQKExhJbnRlcm5ldCBXaWRnaXRzIFB0eSBMdGSCCQDo
oR65d4s5vDAMBgNVHRMEBTADAQH/MA0GCSqGSIb3DQEBBQUAA4ICAQA/Q1W5BBj2
ZobTMge9mwnBh3qNTHpxerokcEjD5Oc65cW10DSTHr/MsAbTUI7t9tTbq1DdUW6D
VpgFglNsaIlKAjvN8WYMDIsNrww5hFXui9JA56sKB6I4cUJMWg0NVhkVB8G7kumX
FaWFmHMcGuxzECT87sP92o8KxpBMHpNtyJuA0lgPrCau0YHM6YkrB9RnUhIjWkGl
DBE8jKrdEmHqwYwPV+DbWbVM3Nnh8y3E6R2VodzOGeedx9b1NnKYU3I8nz9XrAis
frmvezr0+pTbkeM1t+H/oLjAw2v6Zvqn/xsAgRc0G3jpk0C5fYTl8eAUX7Hz9WHq
1UBuDorUOA2NiH7M0NrnDyIJp6l8aJnNGIGTJddOzPvnXFMH1Cafp0JmrK7dshwV
gKMiXogNvW3wy0DJhLOPeACjZEyW4DOawZm3dQJNELWCYa3PrAU4BOS7kyJacPnJ
LtDYOcgcH9dDAZcNsEA9eq713CvtaMdFiuWc+j+JJBZrOVM3wdaTdNqM8++FH8bk
JLd9BjvYUKvsu5GIge1thQxs2GpJC9uf86u9QS60Q1HowY6g0hvw925fUGQ1Jaht
uIBanpU49Fa50IPdD/lwydJzVAorozBvc712u+OtEey+t6AB/9H7aPr8BgdK8sS9
iHYQnid63DkSXnjeoNmjG1OpMzMuRw23rQ==
-----END CERTIFICATE-----"""

with open(APPLIANCE_PUBLIC_KEY_PATH) as appliance_public_key_file:
  APPLIANCE_PUBLIC_KEY = appliance_public_key_file.read().strip()
with open(APPLIANCE_PRIVATE_KEY_PATH) as appliance_private_key_file:
  APPLIANCE_PRIVATE_KEY = appliance_private_key_file.read().strip()


def generate_cert():
  validityEndInSeconds=10 * 365 * 24 * 60 * 60
  k = crypto.PKey()
  k.generate_key(crypto.TYPE_RSA, 4096)
  cert = crypto.X509()
  cert.get_subject().C = "US"
  cert.get_subject().ST = "California"
  cert.get_subject().L = "Sunnyvale"
  cert.get_subject().O = "Fortinet"
  cert.get_subject().OU = "FortiSOAR"
  cert.get_subject().CN = "fsr-sso"
  cert.get_subject().emailAddress = "support@fortinet.net"
  cert.gmtime_adj_notBefore(0)
  cert.gmtime_adj_notAfter(validityEndInSeconds)
  cert.set_issuer(cert.get_subject())
  cert.set_pubkey(k)
  cert.sign(k, 'sha512')
  return crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"), crypto.dump_privatekey(crypto.FILETYPE_PEM, k).decode("utf-8")

def get_public_key_from_cert(cert):
  crtObj = crypto.load_certificate(crypto.FILETYPE_PEM, cert)
  pubKeyObject = crtObj.get_pubkey()
  pubKeyString = crypto.dump_publickey(crypto.FILETYPE_PEM,pubKeyObject)
  return pubKeyString.decode('utf-8')

def regenerate_sso_cert(propagate_to_cluster = False):
    cert, private_key = generate_cert()

    sso_url = "https://localhost/api/saml/config"
    hmac_token = HmacAuth(sso_url, 'GET', APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY)
    response_obj = requests.get(sso_url, auth=hmac_token, verify=False)
    if response_obj.status_code != 200:
        print("SSO settings get failed with status code: " + str(
            response_obj.status_code) + " and error: " + response_obj.text)
        exit(0)
    saml_settings = response_obj.json()
    old_cert = saml_settings["spSettings"]["spX509Cert"]
    if SSO_DEFAULT_CERT != old_cert:
        print("Default SSO certificate can be changed only once")
        exit(0)
    saml_settings["spSettings"]["spX509Cert"] = cert
    saml_settings["spSettings"]["spPublicKey"] = get_public_key_from_cert(cert)
    saml_settings["spSettings"]["spPrivateKey"] = private_key

    hmac_token = HmacAuth(sso_url, 'POST', APPLIANCE_PUBLIC_KEY, APPLIANCE_PRIVATE_KEY, json.dumps(saml_settings))
    response_obj = requests.post(sso_url, json=saml_settings, auth=hmac_token, verify=False)
    if response_obj.status_code == 200:
        print("SSO Certificate updated successfully")
    else:
        print("SSO Certificate update failed with status_code: " + str(response_obj.status_code))


def generate_appliance_keys(propagate_to_cluster = True):
    CSAPIKeys().refresh_keys(uuid=PLAYBOOK_APPLIANCE_UUID)
    postman_uuid = CSUserInfo().get_postman_appliance_key_uuid()
    if postman_uuid:
        CSAPIKeys().refresh_keys(uuid=postman_uuid)


def generate_app_keys(propagate_to_cluster, app=None):
    keys_dir = '/etc/pki/cyops/'
    if not app:
        keys_to_generate = ["das", "wf", "jwt", "postman", "integrations", "gateway", "archival"]
    else:
        keys_to_generate = [app]

    makedirs(keys_dir, exist_ok=True)

    if propagate_to_cluster:
        app_keys = {}
        for key_type in keys_to_generate:
            private_key, public_key = CSAPIKeys().generate_key_pairs()
            write_to_file(public_key, join(keys_dir, key_type+'public.key'))
            write_to_file(private_key, join(keys_dir, key_type+'private.key'))
            app_keys[key_type] = {}
            app_keys[key_type]['public_key'], app_keys[key_type]['private_key'] = public_key, private_key
        save_app_keys_for_new_nodes(app_keys)
        add_cluster_command('generate-app-keys',
                            '/opt/cyops-auth/.env/bin/python /opt/cyops-auth/defaultappliance.py --no-propagate generate-app-keys')
    else:
        app_keys = get_app_keys()
        for key_type in keys_to_generate:
            private_key, public_key = app_keys[key_type]['private_key'], app_keys[key_type]['public_key']
            write_to_file(public_key, join(keys_dir, key_type + 'public.key'))
            write_to_file(private_key, join(keys_dir, key_type + 'private.key'))


def take_arguments():
    parser = argparse.ArgumentParser(description='Utility to generate appliance and app keys')

    subparsers = parser.add_subparsers(help='commands')

    appliance_keys_parser = subparsers.add_parser('generate-appliance-keys', help='Generate Appliance Keys')
    appliance_keys_parser.set_defaults(func=generate_appliance_keys)

    app_keys_parser = subparsers.add_parser('generate-app-keys', help='Generate App Keys i.e. das, jwt, wf and postman')
    app_keys_parser.set_defaults(func=generate_app_keys)

    parser.add_argument('--no-propagate', dest='propagate_to_cluster', action = 'store_false', help = 'Do no propagate the command to cluster')
    parser.add_argument('--app', dest='app', action = 'store', help = 'Any specific app to generate keys')

    sso_cert_parser = subparsers.add_parser('refresh-sso-cert', help='Refresh SSO certificates', usage=argparse.SUPPRESS)
    sso_cert_parser.set_defaults(func=regenerate_sso_cert)

    args = parser.parse_args(args=None if sys.argv[1:] else ['--help'])
    return args


if __name__ == '__main__':
    args = take_arguments()

    if hasattr(args, 'func'):
        if args.func.__name__ == 'generate_app_keys':
            args.func(args.propagate_to_cluster, args.app)
        else:
            args.func(args.propagate_to_cluster)
