#!/opt/cyops-auth/.env/bin/python
""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import argparse, sys, socket, json, os
import signal

if os.getuid() != 0:
    print("This script must be run as 'root'")
    sys.exit(1)

def signal_handler(signal, frame):
    sys.exit(1)

signal.signal(signal.SIGINT, signal_handler)

sys.path.append("/opt/cyops-auth")
from utilities.ha.conf import read_config_file
from utilities.ha.handler import CyOpsHAHandler, refresh_config
from utilities.ha.message_constants import *
from utilities.csconstants import CSConstants, FILE_HANDLER, API_CALL_HANDLER, MESSAGE_HANDLER
from utilities.ha.common_utils import validate_file, print_dataset_with_star_for_current_node, CustomHelpFormatter, print_error, print_dataset, take_yes_no_input
from handlerworkers.cluster import add_sync_cluster_command, get_cluster_commands, add_cluster_command, add_cluster_command_new, add_sync_cluster_command_new
from utilities.ha.consts import CONF_FILE, POSTGRES_SECTION, PG_EXTERNAL, PG_HOST
from utilities.ha.postgres import verify_postgresql_credentials
from log.cslog import get_ha_logger
import utilities.ha.stat
import utilities.ha.health
sys.path.append("/opt/cyops/python_packages/fortisoar")
from fsr.commonlibs.utils import get_list_appliance_flavor
from fsr.commonlibs.consts import APPLIANCE_FLAVOR_FORTICLOUD
from utilities.ha.commands import validator
from utilities.ha.common_utils import decode_command_payload
from utilities.ha.commands.payload_builder import build_cmd_payload
import utilities.ha.slots
logger = get_ha_logger(__name__)

def add_replication_slot_subparser(subparsers_replication_slot):
    parser_list = subparsers_replication_slot.add_parser('list',
                                                          help="List physical replication slots")
    parser_list.set_defaults(operation='utils-replication-slot-list')

    parser_remove = subparsers_replication_slot.add_parser('remove',
                                                            help="Remove physical replication slot")
    parser_remove.set_defaults(operation='utils-replication-slot-remove')
    parser_remove.add_argument('--slot-name',
                               type=str,
                               required=True,
                               dest='slot_name',
                               default=None,
                               help='Physical slot name')

    parser_remove.add_argument('--no-interaction',
                               action='store_false',
                               dest='interactive',
                               default=True,
                               help="Don't ask for any confirmation")

def add_utils_subparser(subparsers):
    parser_utils = subparsers.add_parser('utils',
                                         help="Utilities")
    subparsers_utils = parser_utils.add_subparsers()
    parser_replication_slot = subparsers_utils.add_parser('replication-slots',
                                                          help="Manage physical replication slots")
    subparsers_replication_slot = parser_replication_slot.add_subparsers()
    add_replication_slot_subparser(subparsers_replication_slot)

    for name, subp in subparsers_utils.choices.items():
        subp._positionals.title = 'subcommand options are'
        subp._optionals.title = 'subcommand options are'

def add_suspend_cluster_subparser(subparsers):
    parser_suspend = subparsers.add_parser('suspend-cluster',
                                            help="Suspend cluster for an upgrade")

    parser_suspend.set_defaults(operation='suspend-cluster')
    parser_suspend.add_argument('--no-interaction',
                                action='store_false',
                                dest='interactive',
                                default=True,
                                help="Don't ask for any confirmation")

def add_resume_cluster_subparser(subparsers):
    parser_resume = subparsers.add_parser('resume-cluster',
                                          help="Join back to primary after an upgrade")

    parser_resume.set_defaults(operation='resume-cluster')

def add_clone_db_subparser(subparsers):
    parser_clone_db = subparsers.add_parser('clone-db',
                                            help="Clone a database from the cluster's primary node")

    parser_clone_db.set_defaults(operation='clone-db')

    parser_clone_db.add_argument('--primary-node',
                                action='store',
                                type=str,
                                required= True,
                                dest='primary_node_hostname',
                                default=None,
                                help='Primary node DNS name')

    parser_clone_db.add_argument('--primary-node-ssh-key',
                                action='store',
                                type=str,
                                required=False,
                                dest='primary_node_ssh_key_path',
                                default=None,
                                help='SSH private key of primary node')

def add_joincluster_subparser(subparsers, restricted_join_cluster=False):
    status_choices = ['active', 'passive']
    role_choices = ['primary', 'secondary']

    if restricted_join_cluster:
        status_choices = ['active']
        role_choices = ['primary']

    parser_join_cluster = subparsers.add_parser('join-cluster', help='Join the HA cluster')

    parser_join_cluster.set_defaults(operation='join-cluster')

    parser_join_cluster.add_argument('--status', action='store', type=str, choices=status_choices, required=True)

    # --role required only in case of --status 'active'
    parser_join_cluster.add_argument('--role', action='store', type=str, choices=role_choices,
                                     required='--status' in sys.argv and 'active' in sys.argv,
                                     default='primary',
                                     help='Required only if status is "active"')

    # Node name will be default to hostname
    curr_hostname = socket.gethostname()
    parser_join_cluster.add_argument('--node-name',
                                     action='store',
                                     type=str,
                                     nargs='?',
                                     const=1,
                                     default=curr_hostname,
                                     help='By default set as the current hostname')


    if restricted_join_cluster:
        return

    parser_join_cluster.add_argument('--conf', action='store', type=str, dest='config_file_path',
                                     default=None,
                                     help='Conf file copied from primary server')

    parser_join_cluster.add_argument('--primary-node', action='store', type=str, dest='primary_node_hostname',
                                     default=None,
                                     help='Primary node DNS name')

    parser_join_cluster.add_argument('--primary-node-ssh-key', action='store', type=str, dest='primary_node_ssh_key_path',
                                     default=None,
                                     help='SSH private key of primary node')

    parser_join_cluster.add_argument('--fetch-fresh-backup', action='store_true', dest='fetchfreshbackup', help='Take a fresh backup from primary')
    parser_join_cluster.add_argument('--no-replication-slot',
                                     action='store_false',
                                     dest='use_replication_slot',
                                     default=True,
                                     help=argparse.SUPPRESS)

    parser_join_cluster.add_argument('--skip-local-backup',
                                     action='store_true',
                                     dest='skip_local_backup',
                                     default=False,
                                     help="Skip local database backup. Useful if disk space is low")

    parser_join_cluster.add_argument('--join-cluster-over-api',
                                     action='store_true',
                                     dest='is_join_cluster_over_api',
                                     default=False,
                                     help=argparse.SUPPRESS)

def add_get_config_subparser(subparsers):
    parser_getconf = subparsers.add_parser('export-conf', help='Export the configuration file')
    parser_getconf.set_defaults(operation='export-conf')

def add_whitelist_subparser(subparsers):
    parser_register_server = subparsers.add_parser('allowlist', help="Add secondary/passive server in the allowlist")
    parser_register_server.set_defaults(operation="whitelist")
    parser_register_server.add_argument("--nodes", action='store', type=str, help="Comma-separated list of hostnames of the cluster nodes", required=True)

def add_list_nodes_subparser(subparsers):
    parser_listnodes = subparsers.add_parser('list-nodes', help="List HA cluster details")
    parser_listnodes.set_defaults(operation="list-nodes")
    parser_listnodes.add_argument('--active', action='store_true', dest='is_active', default=None, help='List all active nodes')
    parser_listnodes.add_argument('--passive', action='store_false', dest='is_active', default=None, help='List all passive nodes')
    parser_listnodes.add_argument('--primary', action='store_true', dest='is_primary', default=None, help='List all primary nodes in cluster')
    parser_listnodes.add_argument('--secondary', action='store_false', dest='is_primary', default=None, help='List all secondary nodes in cluster')
    parser_listnodes.add_argument('--count', action='store_true', dest='is_get_count', default=False, help=argparse.SUPPRESS)

def add_leave_cluster_subparser(subparsers):
    parser_leavecluster = subparsers.add_parser('leave-cluster', help="Leave HA cluster")
    parser_leavecluster.set_defaults(operation="leave-cluster")
    parser_leavecluster.add_argument('--no-interaction', action='store_false', dest='interactive', default=True,
                                     help="Don't ask for any confirmation")

def add_list_command_subparser(subparsers):
    listcommands = subparsers.add_parser('list-commands')
    listcommands.set_defaults(func=listcommand_impl, operation='list-commands')
    listcommands.add_argument('--status', action='store', type=str, required=False)
    listcommands.add_argument('--nodeId', action='store', type=str, required=False)

def add_add_command_subparser(subparsers):
    valid_handlers = [FILE_HANDLER, API_CALL_HANDLER, MESSAGE_HANDLER]
    addcommandsync = subparsers.add_parser('add-command')
    addcommandsync.set_defaults(func=addcommand_impl, operation='add-command')
    addcommandsync.add_argument('--async', action='store_true', required=False)
    addcommandsync.add_argument('--handler-type',
                                action='store',
                                type=str,
                                dest='handler_type',
                                choices=valid_handlers,
                                default=None,
                                help='Handler Type')

    addcommandsync.add_argument('--command-type',
                                action='store',
                                type=str,
                                dest='command_type',
                                default=None,
                                help='Commmand Type')

    addcommandsync.add_argument('--payload',
                                action='store',
                                type=str,
                                dest='payload',
                                default=None,
                                help='Input payload')

    # TODO - Legacy arguments, should be removed in future.(We are moving to handler based replication in cluster)
    addcommandsync.add_argument('--commandId', action='store', type=str, required='--handler-type' not in sys.argv)
    addcommandsync.add_argument('--command', action='store', type=str, required='--handler-type' not in sys.argv)

    addcommandsync.add_argument('--nodeId', action='store', type=str, required='--async' not in sys.argv)
    addcommandsync.add_argument('--timeout_in_mins', action='store', type=int, required='--async' not in sys.argv, default=2)

    # file_handler commands specific parameters
    addcommandsync.add_argument('--services-to-restart', dest='services_to_restart', action='store', type=str)

    # file_handler copy command specific parameters
    addcommandsync.add_argument('--paths', dest='paths', action='store', type=str)

    # file_handler copy command specific parameters
    addcommandsync.add_argument('--exclude', dest='exclude', action='store', type=str)

    # file handler share common files specific parameters
    addcommandsync.add_argument('--target-node-id', dest='target_node_id', action='store', type=str)

    # file handler share common files specific parameters
    addcommandsync.add_argument('--exclude-connectors', dest='exclude_connectors', action='store_true', default=False)

    # file_handler rename command specific parameters
    addcommandsync.add_argument('--source', dest='source', action='store', type=str)
    addcommandsync.add_argument('--target', dest='target', action='store', type=str)

    # message_handler specific parameters
    addcommandsync.add_argument('--message', action='store', dest='message', type=str)
    addcommandsync.add_argument('--routing-key', action='store', dest='routing_key', type=str)
    addcommandsync.add_argument('--exchange', action='store', dest='exchange', type=str)

def add_update_hostname_subparser(subparsers):
    update_hostname_parser = subparsers.add_parser('set-node-name')
    update_hostname_parser.set_defaults(operation='set-node-name')
    update_hostname_parser.add_argument("node_name", action='store', type=str)

def add_get_stat_subparser(subparsers):
    get_stat_parser = subparsers.add_parser('get-replication-stat', help="Get the replication statistics")
    get_stat_parser.set_defaults(operation='get-stat')

def add_show_health_subparser(subparsers):
    show_health_parser = subparsers.add_parser('show-health', help="Show the current node health")
    show_health_parser.add_argument('--all-nodes', action='store_true', dest='show_all_nodes_health', default=False, required=False, help="Show all cluster nodes health")
    show_health_parser.add_argument('--json', action='store_true', dest='show_health_in_json_output', default=False,
                                    required=False, help="Show the health as JSON output")
    show_health_parser.add_argument('--no-local-db-point',
                                    action='store_false',
                                    dest='point_local_db',
                                    default=True,
                                    required=False,
                                    help=argparse.SUPPRESS)

    show_health_parser.set_defaults(operation='show-health')

def add_refresh_config_subparser(subparsers):
    parser_refresh_config = subparsers.add_parser('refresh-config')
    parser_refresh_config.set_defaults(operation="refresh-config")

def add_takeover_subparser(subparsers):
    parser_takeover = subparsers.add_parser('takeover', help="Perform takeover")
    parser_takeover.set_defaults(operation="takeover")
    parser_takeover.add_argument('--no-interaction', action='store_false', dest='interactive', default=True,
                                 help="Don't ask for any confirmation")
    parser_takeover.add_argument('--payload',
                                 dest='takeover_payload',
                                 default=None,
                                 required=False,
                                 help=argparse.SUPPRESS)

def add_firedrill_subparser(subparsers):
    parser_firedrill = subparsers.add_parser('firedrill', help="Test DR")
    parser_firedrill.set_defaults(operation="firedrill")

def add_restore_subparser(subparsers):
    parser_restore = subparsers.add_parser('restore',
                                           help="Restore the server to either passive/secondary after firedrill")
    parser_restore.set_defaults(operation="restore")

def define_parser_common(subparsers, limited_parser=True):
    visible_options = "{join-cluster,list-nodes,show-health}"
    subparsers.metavar = visible_options
    if limited_parser:
        add_joincluster_subparser(subparsers, restricted_join_cluster=True)
    add_list_nodes_subparser(subparsers)
    add_show_health_subparser(subparsers)
    add_refresh_config_subparser(subparsers)
    add_update_hostname_subparser(subparsers)
    add_list_command_subparser(subparsers)
    add_add_command_subparser(subparsers)
    add_utils_subparser(subparsers)

def define_parser_enterprise(subparsers):
    visible_options = "{join-cluster,export-conf,allowlist,list-nodes,leave-cluster,takeover,firedrill,restore,get-replication-stat,show-health}"
    subparsers.metavar = visible_options
    add_joincluster_subparser(subparsers)
    add_get_config_subparser(subparsers)
    add_whitelist_subparser(subparsers)
    add_leave_cluster_subparser(subparsers)
    add_get_stat_subparser(subparsers)
    add_takeover_subparser(subparsers)
    add_firedrill_subparser(subparsers)
    add_restore_subparser(subparsers)
    add_clone_db_subparser(subparsers)
    add_suspend_cluster_subparser(subparsers)
    add_resume_cluster_subparser(subparsers)

def define_parser(parser):
    subparsers = parser.add_subparsers()
    limited_parser = True
    list_appliance_flavor = get_list_appliance_flavor()
    if (list_appliance_flavor is None) or (APPLIANCE_FLAVOR_FORTICLOUD not in list_appliance_flavor):
        limited_parser = False
    define_parser_common(subparsers, limited_parser)
    if not limited_parser:
        define_parser_enterprise(subparsers)
    for name, subp in subparsers.choices.items():
        subp._positionals.title = 'subcommand options are'
        subp._optionals.title = 'subcommand options are'

def listcommand_impl(*args, **kwargs):
    command_args = args[0]
    config = read_config_file(CONF_FILE)
    verify_postgresql_credentials(config.get(POSTGRES_SECTION), True)
    commands = get_cluster_commands(command_args['status'], command_args['nodeId'])
    print_dataset_with_star_for_current_node(commands)

def addcommand_impl(*args, **kwargs):
    command_args = args[0]
    is_async = command_args['async']
    payload = build_cmd_payload(command_args)

    if is_async:
        if command_args['handler_type'] is None:
            add_cluster_command(command_args['commandId'], command_args['command'], command_args['nodeId'])
        else:
            payload = decode_command_payload(payload)
            is_valid, message = validator.validate_cmd_payload(command_args['handler_type'], command_args['command_type'], payload)
            if not is_valid:
                print(message)
                sys.exit(1)
            add_cluster_command_new(command_args['handler_type'], command_args['command_type'], payload, command_args['nodeId'])
    else:
        if command_args['handler_type'] is None:
            print(json.dumps(add_sync_cluster_command(command_args['commandId'], command_args['command'],
                                              command_args['nodeId'], command_args['timeout_in_mins'])))
        else:
            payload = decode_command_payload(payload)
            is_valid, message = validator.validate_cmd_payload(command_args['handler_type'], command_args['command_type'], payload)
            print(is_valid)
            if not is_valid:
                print(message)
                sys.exit(1)
            print(json.dumps(add_sync_cluster_command_new(command_args['handler_type'],
                                                          command_args['command_type'],
                                                          payload,
                                                          command_args['nodeId'],
                                                          command_args['timeout_in_mins'])))

def validate_joincluster_inputs(args):
    if args.primary_node_hostname is None and args.config_file_path is None:
        print_error(mconsts_9)
        sys.exit(1)

def start():
    parser = argparse.ArgumentParser(prog="csadm ha", formatter_class=CustomHelpFormatter)

    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'

    define_parser(parser)

    args = parser.parse_args()

    if not 'operation' in dir(args):
        parser.print_help(sys.stderr)
        sys.exit(1)

    if 'config_file_path' in dir(args) and args.config_file_path is not None:
        config_file_path = args.config_file_path.strip()
        validate_file(config_file_path)

    if 'primary_node_ssh_key_path' in dir(args) and args.primary_node_ssh_key_path is not None:
        primary_node_ssh_key_path = args.primary_node_ssh_key_path.strip()
        validate_file(primary_node_ssh_key_path)

    if 'func' in dir(args):
        args.func(vars(args))

    if str(args.operation) == 'join-cluster':
        handler = CyOpsHAHandler()

        if str(args.status) == 'active':
            if str(args.role) == 'primary':
                handler.configure_as_active_primary(args.node_name)
            else:
                config = None
                if args.config_file_path is not None:
                    config = read_config_file(args.config_file_path)
                status = CSConstants.CS_NODE_ACTIVE
                validate_joincluster_inputs(args)
                handler.configure_as_active_secondary(config,
                                                      args.node_name,
                                                      status,
                                                      args.fetchfreshbackup,
                                                      False,
                                                      args.primary_node_hostname,
                                                      args.primary_node_ssh_key_path,
                                                      args.use_replication_slot,
                                                      args.skip_local_backup,
                                                      args.is_join_cluster_over_api)
        else:
            config = None
            if args.config_file_path is not None:
                config = read_config_file(args.config_file_path)
            status = CSConstants.CS_NODE_PASSIVE
            validate_joincluster_inputs(args)
            handler.configure_as_passive_secondary(config,
                                                   args.node_name,
                                                   status,
                                                   args.fetchfreshbackup,
                                                   False,
                                                   args.primary_node_hostname,
                                                   args.primary_node_ssh_key_path,
                                                   args.use_replication_slot,
                                                   args.skip_local_backup,
                                                   args.is_join_cluster_over_api)

        return

    if str(args.operation) == 'export-conf':
        handler = CyOpsHAHandler()
        handler.exportconf()
        return

    if str(args.operation) == 'whitelist':
        handler = CyOpsHAHandler()
        handler.whitelist(args.nodes)
        return
    
    if str(args.operation) == 'list-nodes':
        handler = CyOpsHAHandler()
        handler.get_node_list(args.is_active, args.is_primary, args.is_get_count)
        return

    if str(args.operation) == 'takeover':
        handler = CyOpsHAHandler()
        handler.takeover_as_active_primary(interactive=args.interactive, takeover_payload=args.takeover_payload)
        return
  
    if str(args.operation) == 'restore':
        handler = CyOpsHAHandler()
        handler.restore_after_firedrill()
        return

    if str(args.operation) == 'firedrill':
        handler = CyOpsHAHandler()
        handler.firedrill()
        return 

    if str(args.operation) == 'leave-cluster':
        handler = CyOpsHAHandler()
        handler.leavecluster(args.interactive)
        return

    if str(args.operation) == 'set-node-name':
        handler = CyOpsHAHandler()
        handler.update_node_name(args.node_name)
        return

    if str(args.operation) == 'get-stat':
        utilities.ha.stat.get_stat(print_stat=True)
        return

    if str(args.operation) == "refresh-config":
        refresh_config()
        return

    if str(args.operation) == "show-health":
        utilities.ha.health.print_cluster_health(show_all_nodes_health=args.show_all_nodes_health,
                                                 show_health_in_json_output=args.show_health_in_json_output,
                                                 point_local_db=args.point_local_db)
        return

    if str(args.operation) == "clone-db":
        utilities.ha.postgres.clone_db(**vars(args))
        return

    if str(args.operation) == "suspend-cluster":
        handler = CyOpsHAHandler()
        handler.suspend_cluster(args.interactive)
        return

    if str(args.operation) == "resume-cluster":
        handler = CyOpsHAHandler()
        handler.resume_cluster()
        return

    if str(args.operation) == "utils-replication-slot-list":
        utilities.ha.slots.list_slots()
        return

    if str(args.operation) == "utils-replication-slot-remove":
        current_config = read_config_file()
        current_pg_config = current_config[POSTGRES_SECTION]
        if current_pg_config[PG_EXTERNAL]:
            print(MSG_REMOVE_SLOT_NOT_APPLICABLE)
            return
        # Point to localhost, no need to worry about password, since it will be same for cluster
        current_pg_config[PG_HOST] = 'localhost'
        if args.interactive:
            input_ = take_yes_no_input(MSG_WARN_REMOVE_SLOT)
            if input_.lower() != 'yes':
                return
        utilities.ha.slots.delete_slot(current_pg_config, args.slot_name)

if __name__ == '__main__':
    start()
