#!/opt/cyops-auth/.env/bin/python
""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import os
import sys
import argparse


script_orig_cwd = os.getcwd()
sys.path.append("/opt/cyops-auth")

# Include common packages in licensemanager
sys.path.append("/opt/cyops/python_packages/fortisoar")
from fsr.commonlibs.utils import CustomHelpFormatter


if not os.geteuid() == 0:
    sys.exit("You need to be root to perform this command.")

# Include common packages in licensemanager
from handlerworkers.users import logout_user_sessions, get_uuid_by_loginid, get_logged_in_users
from utilities.csconstants import NAMED_ACCESS_TYPE, CONCURRENT_ACCESS_TYPE


def logout_user(args):
    uuid = get_uuid_by_loginid(args.username)
    if not uuid:
        print(f"User {args.username} does not exist. Please provide valid username.")
        return
    status, message = logout_user_sessions(uuid, is_force_logout=True, logout_from_ch=True, is_cli_logout=True)
    print(message)


def show_logged_in_users(args):
    users = get_logged_in_users(access_type=args.access_type, limit=args.limit)
    if not users:
        print("No users logged in currently")
    else:
        print("Users with following usernames are logged in currently")
        print(*users, sep=", ")


def define_parser(parser):
    subparsers = parser.add_subparsers()
    show_logged_in_users_subparser = subparsers.add_parser('show-logged-in-users', help="Show logged in users")
    show_logged_in_users_subparser.set_defaults(func=show_logged_in_users, operation='show-logged-in-users')
    show_logged_in_users_subparser.add_argument("--access-type", choices=[NAMED_ACCESS_TYPE, CONCURRENT_ACCESS_TYPE],
                                                dest='access_type', default=CONCURRENT_ACCESS_TYPE,
                                                help='User access type. Default Concurrent')
    # range function in python does not include the last number
    show_logged_in_users_subparser.add_argument("--limit", type=int, dest='limit', default=10, choices=range(1,31),
                                                metavar="[1-30]", help='Number of users to show. Default 10')

    logout_user_subparser = subparsers.add_parser('logout-user', help="Logout a user")
    logout_user_subparser.set_defaults(func=logout_user, operation='logout-user')
    logout_user_subparser.add_argument("--username", dest='username', required=True,
                                       help='Username of user to be logged out')

def start():
    parser = argparse.ArgumentParser(prog="csadm user", formatter_class=CustomHelpFormatter)
    define_parser(parser)
    parser._positionals.title = 'subcommand options are'
    parser._optionals.title = 'subcommand options are'
    args = parser.parse_args()

    if args.operation == 'show-logged-in-users':
        show_logged_in_users(args)
        return
    if args.operation == 'logout-user':
        logout_user(args)
        return


if __name__ == "__main__":
    start()
