""" Copyright start
  Copyright (C) 2008 - 2022 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
import re
from calendar import timegm
from datetime import datetime

from handlerworkers.configprops import CSConfigProps
from log.cslog import get_logger
from utilities.csbcrypt import CSBcrypt
from validationutils.abstractpwdvalidator import PasswordValidator

logger = get_logger(__name__)

class CustomPasswordValidator(PasswordValidator):
    ##########################################
    #
    # validate
    #
    ##########################################
    def validate(self, password, loginid=None, password_history=None):


        """
        This function simply calls other functions to validate various
        aspects of the password.
        :param password:
        :return: True/False
        """
        if password_history:
            for password_detail in password_history:
                if CSBcrypt().is_valid_password(password_detail[1], password):
                    message = "Password matches with one of your previous passwords"
                    logger.error(message)
                    return False, message

            last_password_detail = password_history[-1]
            last_password_changed_at = last_password_detail[0]
            now = timegm(datetime.utcnow().utctimetuple())
            min_password_age = CSConfigProps().get_section('PASSWORD').get('min_password_age')
            if now-min_password_age*24*60*60 <= last_password_changed_at:
                message = "Password change not allowed within "+str(min_password_age)+" day(s) of last password change"
                logger.error(message)
                return False, message

        if loginid.lower() in password.lower():
            message = "Username not allowed in password"
            logger.error(message)
            return False, message

        assert isinstance(password, str)
        if len(password) < 8:
            message = "Password must be at least 8 character long"
            logger.error(message)
            return False, message
        if not re.search("[a-z]", password):
            message = "Password must contain at least one small letter"
            logger.error(message)
            return False, message
        if not re.search("[A-Z]", password):
            message = "Password must contain at least one capital letter"
            logger.error(message)
            return False, message
        if not re.search("[0-9]", password):
            message = "Password must contain at least one digit"
            logger.error(message)
            return False, message
        if not re.search("[~!@#$%^&*|?_]", password):
            message = "Password must contain one special character from #$%&?@_"
            logger.error(message)
            return False, message
        if re.search("\s", password):
            message = "Password must not contain whitespaces"
            logger.error(message)
            return False, message
        return True, "Password validation successful"

    def password_policy(self):
        return ["Your password must have at least 8 characters", "Password must contain one lower-case alphabet, one upper-case alphabet, one digit, and any one of these special characters ~!@#$%^&*|?_",
                "Must not have been used in the previous 10 passwords",
                "Must not have been changed within the last 24 hours", "Must not contain a user’s Account name"]

